/*
 * This file implements a socket wrapper to provide some basic TLS client
 * support using the openssl library.  
 */

/*
 * This file is part of the Cleansocks library.  It is released under the
 * Lesser GNU GPL.  See the accompanying files COPYING and COPYING.LESSER.
 */
#include <cctype>

#include <string>
#include <fstream>

#include <stdlib.h>

#include <openssl/ssl.h>
#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/x509.h>
#include <openssl/opensslv.h>

#include "cleanbio.h"
#include "cleantlsc.h"

namespace cleansocks {

SSL_CTX *client_tls_socket::sm_ctx = NULL;

// User-provided root cert path
std::string client_tls_socket::sm_authority_path = "";

// Some possibility we found on our own.
std::string client_tls_socket::sm_found_authority =
	client_tls_socket::find_authority();

// This just checks a few standard locations for an authority .pem, and returns
// the first one it finds.  Doesn't try to load the file; just tries to see
// if an open succeeds.  I really want to use the lists from installed browsers,
// but they don't seem to use the .pem format that openssl likes.
std::string client_tls_socket::find_authority()
{
	// Since this method is static, and its value is used to initialize
	// a static field used by start_tls(), these will have to be run
	// before TLS is started.
	SSL_library_init();
	SSL_load_error_strings();

	// Locations to try.  Each location is a list given as a list of strings.  If a string starts
	// with @, the rest is an env variable, which will be followed by a default.  If the variable is
	// not set and the default is NULL, the test is cancelled.  The name is built from the
	// concat of the resulting strings.
#ifdef WIN32
	const char *l1[] = { "@HOMEDRIVE", "C:", "@HOMEPATH", NULL, "\\cleansocks\\ca.pem", NULL };
	const char *l2[] = { "@HOMEDRIVE", "C:", "@HOME", NULL, "\\cleansocks\\ca.pem", NULL };
	const char *l3[] = { "@SystemDrive", "C:", "@ProgramFiles", NULL, "\\cleansocks\\ca.pem", NULL};
	const char *l3b[] = { "@SystemDrive", "C:", "@ProgramFiles(x86)", NULL, "\\cleansocks\\ca.pem", NULL};
	const char *l4[] = { "@SystemDrive", "C:", "\\Program Files\\cleansocks\\ca.pem", NULL};
	const char *l5[] = { "C:\\Program Files\\cleansocks\\ca.pem", NULL};

	const char * const *locs[] = {l1, l1+2, l2, l2+2, l3, l3+2, l3b, l3b+2, l4, l4+2, l5, NULL};
#else
	const char *l1[] = { "@HOME", NULL, "/.cleansocks/ca.pem", NULL };
	const char *l2[] = { "/usr/local/etc/cleansocks/ca.pem", NULL };
	const char *l3[] = { "/etc/cleansocks/ca.pem", NULL };
	const char *l4[] = { "/etc/pki/tls/cert.pem", NULL };  /* At least Fedora puts it here. */
	const char *l5[] = { "/etc/pki/ca-trust/extracted/pem/tls-ca-bundle.pem", NULL };

	const char * const *locs[] = {l1, l2, l3, l4, l5, NULL};
#endif
	// Run through the list and look for the first one that can be read.  
	for(const char * const **scan = locs; *scan; ++scan) {
		// Run through the list of parts denoted by scan and build the 
		// result in path.
		std::string path("");
		for(const char * const *pscan = *scan; *pscan; ++pscan) {
			// Get the part.  If it starts with @, replace with the environment
			// lookup.
			const char *part = *pscan;
			if(part[0] == '@') {
				// Skip the @EV spec in the list, and point to the default.
				++pscan;

				// Get the environment variable.  If that fails, replace
				// with the default.  If that is also NULL, clear path
				// and bail.  The location will not be checked.
				part = getenv(part+1);
				if(part == NULL) {
					part = *pscan;
					if(part == NULL) {
						path = "";
						break;
					}
				}
			}

			path += part;
		}
		if(path == "") continue;

		// See if we can open the file, read a byte, and see if its printable.  I'd 
		// like to just do an open, but [insert earned insult] Windows doesn't
		// seem to fail on missing files based on just an open.  I suppose it
		// might be a mingw problem, actually.  Wrappers do like to delay stuff.
		// (Hope to heck it manages to close it properly on return.)
		std::ifstream ifs(path.c_str());
		char ch = ifs.get();
		if(ifs && isascii(ch) && (isprint(ch) || ch == '\n' || ch == '\r'))
			return path;
	}
	return "";
}

void client_tls_socket::allocate_context()
{
	if(sm_ctx == NULL) {
#if OPENSSL_VERSION_NUMBER >= 0x10100070L
		sm_ctx = SSL_CTX_new(TLS_client_method());
#else
		sm_ctx = SSL_CTX_new(SSLv23_client_method());
#endif
		if(sm_ctx == NULL)
			throw socket_tls_context_create();
		if(sm_authority_path != "") {
			if(SSL_CTX_load_verify_locations
			   (sm_ctx, sm_authority_path.c_str(), 
			    sm_authority_path.c_str()) <= 0)
				throw socket_tls_set_ca(sm_authority_path);
		} else if(sm_found_authority != "") {
			if(SSL_CTX_load_verify_locations
			   (sm_ctx, sm_found_authority.c_str(), NULL) <= 0)
				throw socket_tls_set_ca(sm_found_authority);
		} else 
			throw socket_tls_ca_path();
		if(m_certhost != "" && !(m_flags & NO_HOST_CHK)) {
			// Add the host check to the automatic verification.
			X509_VERIFY_PARAM *param = X509_VERIFY_PARAM_new();
			X509_VERIFY_PARAM_set1_host(param, m_certhost.data(),
						    m_certhost.length());
			SSL_CTX_set1_param(sm_ctx, param);
			X509_VERIFY_PARAM_free(param);			
		}
	}
}

void client_tls_socket::start_tls_private(client_tls_socket *share_sock)
{
	// Create and verify the SSL connection.  If session sharing is
	// specified, tell OpenSSL about it.
	if(m_ssl != NULL) throw socket_tls_start_started();
	if((m_ssl = SSL_new(sm_ctx)) == NULL) throw socket_tls_ssl_create();
#if OPENSSL_VERSION_NUMBER >= 0x10000000L
	if(m_certhost != ""  && !(m_flags & NO_SNI)) {
		if(!SSL_set_tlsext_host_name(m_ssl, m_certhost.c_str()))
			throw socket_tls_unable_to_set_sni_host(m_certhost);
	}
#endif
	SSL_set_mode(m_ssl, SSL_MODE_AUTO_RETRY|SSL_MODE_ENABLE_PARTIAL_WRITE);
	BIO *bio = BIO_new_clean(m_base, BIO_NOCLOSE);
	if(bio == NULL) throw socket_tls_creating_bio();
	SSL_set_bio(m_ssl,bio,bio);
	SSL_set_connect_state(m_ssl);
	if(share_sock) {
		// If requested, configure session reuse.
		if(share_sock->m_ssl == NULL)
			throw socket_tls_session_sharing_fail
				("No session to share");
		SSL_SESSION *sess = SSL_get_session(share_sock->m_ssl);
		if(sess == NULL)
			throw socket_tls_session_sharing_fail();
		if(!SSL_set_session(m_ssl, sess))
			throw socket_tls_session_sharing_fail();
	}
	if(SSL_connect(m_ssl) <= 0) throw socket_tls_handshake();
	if(m_flags & NO_CERT_CHECK) return;

	// Get the cert results.
	long vcode = SSL_get_verify_result(m_ssl);
#if OPENSSL_VERSION_NUMBER >= 0x30000000L
	X509 *cert = SSL_get1_peer_certificate(m_ssl);
#else
	X509 *cert = SSL_get_peer_certificate(m_ssl);
#endif	
	if(cert != NULL) {
		X509_free(cert);
	}

	// Certain types of errors are forgiven.
	switch(vcode) {
	case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
	case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
		if(m_flags & SELF_OK) vcode = X509_V_OK;
		break;
	case X509_V_ERR_CRL_HAS_EXPIRED:
		if(m_flags & EXPIRED_OK) vcode = X509_V_OK;
		break;
	}

	// Bail on error.
	if(vcode != X509_V_OK || (cert == NULL && !(m_flags & NO_CERT_OK))) {
		stop_tls();
		throw socket_tls_bad_cert(vcode);
	}
}
int client_tls_socket::send(const void *buf, int size)
{
	if(m_ssl) {
		if(size == 0) return 0;
		int ret = SSL_write(m_ssl, buf, size);
		if(ret == 0 && SSL_get_error(m_ssl, ret) == 
		   SSL_RECEIVED_SHUTDOWN)
			throw socket_tls_send_shut();
		if(ret <= 0) throw socket_tls_send();
		return ret;
	} else
		return cleansocks::send(*m_base,buf,size);
}
int client_tls_socket::recv(void *buf, int size)
{
	if(m_ssl) {
		int ret = SSL_read(m_ssl, buf, size);
		if(ret <= 0) {
			int err = SSL_get_error(m_ssl, ret);
			ret = 0;
			if(err != SSL_ERROR_NONE && 
			   err != SSL_ERROR_ZERO_RETURN &&
			   err != SSL_RECEIVED_SHUTDOWN &&
#ifdef SSL_R_UNEXPECTED_EOF_WHILE_READING 			   
			   err != SSL_R_UNEXPECTED_EOF_WHILE_READING &&
#endif			   
			   !(err == SSL_ERROR_SYSCALL && errno == 0))
				ret = -1;
			//abort();
		}
		if(ret < 0) throw socket_tls_recv();
		return ret;
	} else
		return cleansocks::recv(*m_base,buf,size);
	
}
void client_tls_socket::stop_tls()
{
	if(m_ssl == NULL) return;
	int sd = SSL_shutdown(m_ssl);
	if(sd == 0) sd = SSL_shutdown(m_ssl);
	//if(sd < 0) { ... } // Keep getting sd < 0 and no error mesage.
	SSL_free(m_ssl);
	m_ssl = NULL;
}

// Usually get the TLS error from the system.  Which seems to be a total mess.
std::string socket_tls_error::get_err(std::string msg) {
	int err = ERR_peek_error();
	if(err != 0) {
		char buf[120];
		ERR_error_string(err, buf);
		msg += std::string(": ") + buf;
	}
	return msg;
}

// Print the full error stack.
void socket_tls_error::error_stack() {
	std::cerr.flush();
	ERR_print_errors_fp(stderr);
}

// Return a string form of the cert verification code
#include "cert_err_codes.inc"
std::string socket_tls_bad_cert::err_code(long code)
{
	if(code < 0 || code > cert_err_last || cert_err_list[code] == NULL)
		return "Unidentified reason";
	return
		cert_err_list[code];
}

}
