/*
 * This is a test of the buffered socket designed to make sure it can
 * drain all the available bytes.  (I've seen some evidence of a bug here,
 * want to create a test that exercises it.)
 *
 * It simply writes some bytes 127.0.0.1, then reads them out again using
 * the buffered reader.  Counts are parameters.
 *
 * pgm [ -v ] [ -b bufsize ] [ -p port ] [ -c ] numwrite read1 [ t ] [ a ] .... [ x | d | t ] [ a ] ]
 *   Write the numwrite number of bytes, then perform reads of each number.
 *   Writes are of the upper-case alphabet, repeating as needed.  Reads are
 *   performed regardless, even if reading exceeds the number of writes, or if
 *   a read is zero.  If the last read parameter is x, and there are some
 *   remaining unrequested bytes, issue a final read of the rest of them.  
 *   The d is the same, using the count of unreceived bytes.  Reads are with
 *   plain read, unless followed by an uppercase letter, in which case readln
 *   with that letter terminator is used.  If the read is followed by a
 *   lower-case a, the WAITALL flag is sent.  The -b introduces a bufsize
 *   other than the default, and -v indicates that the received data should
 *   be printed.  (It is checked in any case.)  The -p specifies a port number
 *   on localhost to use.  The -c option means the sender should close its
 *   socket after sending.  Otherwise, it stays open.
 */	
/*
 * This file is part of the Cleansocks library.  It is released under the
 * Lesser GNU GPL.  See the accompanying files COPYING and COPYING.LESSER.
 */

#include <cstdlib>
#include <cctype>
	
#include <iostream>
#include <string>
#include <stdexcept>
#include <thread>
#include <mutex>

#include "cleansocks.h"
#include "cleanip.h"
#include "cleanbuf.h"

using std::string;
using std::cout;
using std::cerr;
using std::endl;
using namespace cleansocks;

string pname;

#ifdef WIN32
// From docs.  May never actually try to compile it on Windows.
#include <Processthreadsapi.h>
inline int getpid() { return (int)GetCurrentProcessId(); }
#endif

// Data sending thread.
std::mutex close_control;
void sendthread(int numwrite, int port)
{
	TCPsocket sendsoc;
	connect(sendsoc, IPendpoint(IPaddress("127.0.0.1"),port));

	const char *alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
	while(numwrite > 0) {
		int tosend = numwrite;
		if(tosend > 26) tosend = 26;
		if(send(sendsoc, alphabet, tosend) != tosend)
			throw socket_error("Short write");
		numwrite -= tosend;
	}

	// Prob'ly should use an event, but this works well enough.
	close_control.lock();
	close(sendsoc);
	close_control.unlock();
}

int main(int argc, char **argv)
{
	pname = *argv++;
	--argc;

	int bufsize = buffered_socket::DEF_BUF_SIZE;
	bool verbose = false;
	int port = (getpid() % 10000) + 40000;
	bool doclose;
	for(;argc > 0 && argv[0][0] == '-'; ++argv, --argc) {
		if(string(*argv) == "-b" && argc > 1) {
			bufsize = atoi(*++argv); --argc;
		} else if(string(*argv) == "-p" && argc > 1) {
			port = atoi(*++argv); --argc;
		} else if(string(*argv) == "-v") {
			verbose = true;
		} else if(string(*argv) == "-c") {
			doclose = true;
		} else {
			std::cerr << "Bad flag " << *argv << std::endl;
			exit(1);
		}
	}

	// Listen on the local address.
	TCPsocket listenloc;
	bind(listenloc, IPendpoint(IPaddress("127.0.0.1"), port));
	listen(listenloc);
	
	// Get the sending size.
	int numwrite = 100;
	if(argc > 0) {
		numwrite = atoi(*argv++);
		--argc;
	}
	int pending = numwrite;  // Sent but not received

	// If we want to prevent close, nab the close lock to keep the
	// child thread from getting it.
	if(!doclose) close_control.lock();
	
	// Start the sender.
	std::thread sender(sendthread, numwrite, port);

	// Get the receving socket.
	TCPsocket rcver = accept(listenloc);
	buffered_socket brecv(rcver, bufsize);

	// Some giant buffer for receives.
	static const int RBSIZE = 10*1024;
	char rbuf[RBSIZE];

	// Proeccess receive instructions
	char nextlet = 'A';	// Next byte to receive.
	int count = 1;		// Count of operations.
	while(argc > 0) {
		// Find the number
		int numread;
		if(string(*argv) == "x") {
			if(numwrite <= 0) break;
			numread = numwrite;
			++argv; --argc;
		} else if(string(*argv) == "d") {
			if(pending <= 0) break;
			numread = pending;
			++argv; --argc;
		} else {
			numread = atoi(*argv++); --argc;
			if(numread < 0) {
				std::cerr << "Negative read size " << numread
					  << " ignored." << std::endl;
				continue;
			}
		}
		numwrite -= numread;

		// Note the flags.
		int term = -1;
		clean_flags flags(0);
		for(; argc > 0; ++argv, --argc) {
			if(isupper(**argv)) {
				term = **argv;
				continue;
			} else if(string(*argv) == "a") {
				flags = CSMSG_WAITALL;
				continue;
			}
			break;
		}

		// Sanity check here.
		if(numread > RBSIZE) {
			std::cerr << "Sorry, read size of " << numread
				  << " exceeds buffer size " << RBSIZE
				  << std::endl;
			exit(3);
		}

		// Perform the read.
		int nread = recvln(brecv, rbuf, numread, term, flags);
		pending -= nread;
		if(verbose) {
			cout << "read " << nread << " "
			     << string(rbuf, nread) << endl;
		}
		int errct = 0;
		int first = -1;
		for(int i = 0; i < nread; ++i) {
			if(rbuf[i] != nextlet) {
				if(first == -1) first = i;
				++errct;
				nextlet = rbuf[i];
			}
			nextlet = nextlet == 'Z' ? 'A' : nextlet + 1;
		}
		if(errct) {
			cout << "Receive " << count << " wrong at " << first
			     << ", total " << errct << " errors." << endl;
		}

		if(nread != numread) {
			if(flags == CSMSG_WAITALL)
				cout << "Short waitall receive: " << nread
				     << " instead of " << numread << "."
				     << endl;
			else if(verbose)
				cout << "Note: short receive: " << nread
				     << " < " << numread << endl;
		}

		++count;
	}

	// If we acquired the lock to prevent close, let him have it now.
	// so he closes the sending socket.
	if(!doclose) close_control.unlock();
	sender.join();
}
