/*
 * FTP directory listing program.
 *
 *   $0 [acct@]server [ path ]
 *
 * This will log in to the server.  If an acct is given, the program will
 * log in using those credentials, otherwise as anonymous.  If the path is
 * given, the program will move to the indicated directory.  It will then
 * request a listing of the directory from the server, and print that listing
 * to the console.  If an account is given, the program will ask for a
 * password before logging into the server.  If the login fails, it will
 * ask for the password again.
 *
 * https://tools.ietf.org/html/rfc4217
 */

#include <iostream>
#include <stdexcept>
#include <fstream>
#include <sstream>
#include <string>
#include <array>

#include <stdlib.h>
#include <string.h>

// Use either the Unix system getpass, or the one provided with the assignment,
// depending on where we're compiling.
#ifdef _WIN32
#include "getpass.h"
#else
#include <unistd.h>
#endif

using namespace std;

#include <cleansocks.h>
#include <cleanip.h>
#include <cleanbuf.h>
#include <cleantlsc.h>

using namespace cleansocks;

// Markers for the two sides of the conversation.
static const char *server_mark = "SVR: ";
static const char *client_mark = "CLI: ";
static const char *note_mark = ">>>> ";

// Print paramter usage and exit.
static char *pname;
void usage()
{
	cerr << pname << " [ -p | -s | -S ] [ -P port ] [acct@]server [ path ]"
	     << endl;
	exit(1);
}

// It's convenient here to have an exception representing a error reported to
// us by the server, for which we should send QUIT.  This is distinct from
// networking failures, which usually mean that communication with the server
// is impossible, and result in exceptions sent from the cleansocks library.
class bail_out: public runtime_error {
public:
	bail_out(const string &s): runtime_error(s) { }
};

// Read and echo an FTP server response, allowing for continued lines.  Return
// the response code.  NOTE: the FTP protocol (RFC 959) does not specify a
// maximum line length, so we pick 1024, which should be plenty, and puke if
// some rude server tries to send us a line too long for that.
//
// If passinfo is non-NULL, response() checks each line for a substring that
// looks like a PASV response.  If found, it is placed there.  If found in
/// multiple lines (unlikely), the last one will be returned.
int response(buffered_socket &s, string *passinfo = NULL)
{
	// Hold the code part of the first response line.  Used to identify
	// the last line line, per RFC 959.
	const int codesize = 3;
	std::string last;

	// Read all the lines.
	bool firstline = true;
	while(true) {
		// Read a line from the server, and make sure it fits in the
		// buffer.  Then, look for the \r\n ending, and trim the
		// length so the line can be used as a string sans terminator.
		std::array<char, 1024> buf;
		int n = recvln(s, buf);
		if(n >= 2 && buf[n-2] == '\r' && buf[n-1] == '\n') {
			// Fits.  Remove the terminators by reducing size.
			n -= 2;
		}
		else if(n < 2)
			// Should not give such an error.
			throw bail_out("Short/empty response, n = " +
				       to_string(n));
		else
			// The \r\n not found at the end.  Either the response
			// is ill-formed, or the line is just longer than our
			// buffer size.
			throw bail_out("Oversize or ill-formed server "
				       "response");

		// Make the response into a string and echo it.
		string line(buf.begin(),buf.begin()+n);
		cout << server_mark << line << endl;

		// Optionally, extract the PASV response info.
		if(passinfo != NULL) {
			int left = line.find("(");
			if(left != string::npos) {
				++left;
				int right = line.find_first_not_of
					("0123456789,",left);
				if(right != string::npos && line[right] == ')')
					*passinfo =
						line.substr(left,right-left);
			}
		}

		// Special processing for the first (most often only) line
		// of the response.
		if(firstline) {
			// First line.  If the response is not long enough to
			// contain a code and space, it violates protocol.
			if(n <= codesize)
				throw bail_out("Short server response");

			// Record the code, with a space after it.  That's
			// what marks the last line.  Also used to generate
			// the function return value.
			last = string(buf.begin(), buf.begin()+codesize) + " ";

			// If there's not a continue mark, this is the only 
			// line, so we break now.
			if(buf[codesize] != '-') break;

			// Not the first line anymore.
			firstline = false;

			// We'll continue looping until we find the last line.
		} else {
			// Continuation line.  Break if it's the last one.
			if(n > codesize &&
			   last == string(buf.begin(),
					  buf.begin() + codesize + 1))
				break;
		}
	}

	// Convert and return the response code.
	return stoi(last);
}

// Send and echo a command, print the response, and return the code.
int request(buffered_socket &s, const string &req)
{
	cout << client_mark << req << endl;
	send(s, req + "\r\n");
	return response(s);
}

// Send and echo a command, print the response, and throw with the indicated
// message if the response is not a success.
inline void do_or_die(buffered_socket &s, const string &req, const string &msg,
		      int success = 2)
{
	if(request(s, req) / 100 != success)
		throw bail_out(msg);
}

// Perform the AUTH TLS and associated commands to start the TLS
// connection.  Will throw if it doesn't work out.
void do_request_tls(buffered_socket &control, client_tls_socket &tlslayer,
		    string prot) // prot = "P" or "C".
{
	do_or_die(control, "AUTH TLS", "Server will not accept TLS.");
	tlslayer.start_tls(); // Throws on error.
	cout << note_mark << "Control TLS Established" << endl;
	do_or_die(control, "PBSZ 0", "Unable to set security buffer size");
	do_or_die(control, "PROT "+prot, "Unable to select secure connection");
}

// Log in to the server using the pre-existing control connection and the given
// credentials.  Returns true of the login succeeded, false if it ended with
// a code 530, indicating an incorrect login.  Throws for other errors.
bool do_login(buffered_socket &control, string login, string passwd)
{
	// Send the user command to log in.
	int code = request(control, "USER " + login);

	// If the server wants a password send it.  (It pretty much
	// always will.)
	if(code == 331) {
		// Try sending the password.  We do our special echo
		// to suppress printing the password.
		cout << client_mark << "PASS ******" << endl;
		send(control, "PASS " + passwd + "\r\n");
		code = response(control);
	}

	// We should be logged in.  If not, puke.
	if(code == 332)
		throw bail_out("We don't handle accounts");
	else if(code / 100 == 2)
		return true;
	else if(code == 530)
		return false;
	else
		throw bail_out("Login not accepted.");
}

// Try to log in with a password from the user.  Succeeds or throws.
void do_user_login(buffered_socket &control, string login)
{
	char *pass;
	do {
		pass = getpass("Password: ");
		if(pass == NULL)
			throw bail_out("Password read failed");
		if(*pass == '\0')
			throw bail_out("Login failed");
	} while(!do_login(control, login, pass));
} 

// Treat a prefix of the string up to , as a number, remove and return the
// number as an integer.  Used to extract passive response data.
int get_first_int(string &str)
{
	int ret = 0;
	int loc = str.find(",");
	if(loc != string::npos) {
		ret = stoi(str.substr(0,loc));
		str.erase(0,loc+1);
	}

	return ret;
}

// Create an endpoint from the PASV response string.
IPendpoint make_endpoint(string passinfo)
{
	// Extract the port info.
	unsigned long addr = 0;
	for(int i = 0; i < 4; ++i)
		addr = 256*addr + get_first_int(passinfo);
	unsigned short pno = get_first_int(passinfo);
	pno = 256*pno + get_first_int(passinfo);
	return IPendpoint(addr, pno);
}

// Make a passive connection and print a listing of the current directory
// to the console.
void do_listing(buffered_socket &control, client_tls_socket *csock)
{
	// Ask for passive mode, which gives a host and port to connect to.
	string passinfo = "";
	cout << client_mark << "PASV" << endl;	
	send(control, (string)"PASV\r\n");
	int code = response(control, &passinfo);
	if(code != 227 || passinfo == "")
		throw bail_out("PASV request failed.");
	
	// Connect to the specified endpoint.  (Throws on failure.)
	TCPsocket dsock;
	IPendpoint ep = make_endpoint(passinfo+",");
	cout << note_mark << "Connecting to " << ep << endl;
	connect(dsock, ep);
	client_tls_socket dldsock(dsock, NO_CERT_CHECK, false);  

	// Attempt the transfer.  Use a try so the port gets closed after
	// some error.
	try {
		// Request the listing.
		do_or_die(control, "LIST", "Directory listing failed", 1);

		// Turn on TLS if needed.
		if(csock) {
			dldsock.start_tls_share_session(*csock);
			cout << note_mark << "TLS started for data stream"
			     << endl;
		}

		// Copy the contents to the standard out.
		char buf[1024];
		int size;
		while((size = recv(dldsock, buf, sizeof buf)) > 0) {
			cout.write(buf, size);
		}
		close(dldsock);
	} catch(exception &e) {
		close(dldsock);
		throw e;
	}

	// Should be a response message.
	if(response(control) / 100 != 2)
		throw bail_out("Listing failed");
}

int main(int argc, char **argv)
{
	// Save the executable name and remove it from the list.
	pname = *argv++;
	--argc;

	// Flags: -s for explicit starttls (RFC 4217), -S for implicit ftp
	// over TLS.  Also -p to set the control connection port number.
	bool starttls = false;		// Explicit TLS on control.
	bool datacrypt = false;		// Encrypt data stream.
	bool ftpovertls = false;	// Old-style implicit TLS.
	int portno = -1;
	while(argc > 0 && **argv == '-') {
		if(strcmp(*argv, "-s") == 0) {
			datacrypt = starttls = true;
		} else if(strcmp(*argv, "-p") == 0) {
			starttls = true;
		} else if(strcmp(*argv, "-S") == 0) {
			datacrypt = ftpovertls = true;
		} else if(strcmp(*argv, "-P") == 0) {
			if(argc < 1 || argv[1][0] == '-') usage();
			portno = atoi(argv[1]);
			if(portno <= 0 || portno > 0x0ffff) usage();
			--argc; ++argv;
		} else {
			usage();
		}
		--argc; ++argv;
	}
	if(starttls && ftpovertls) usage();

	// Get the argument info.
	if(argc < 1 || argc > 2) usage();

	// Get the server string (or maybe account@server).
	string server = argv[0];

	// Get the account, either default or by trimming from server.
	string login = "anonymous";
	int pos = server.find("@");
	if(pos != string::npos) {
		login = server.substr(0,pos);
		server.erase(0,pos+1);
	}

	// Interact with the server.
	TCPsocket sock;
	client_tls_socket tls;
	buffered_socket control;
	try {
		// Make the control connection.
		IPaddress addr = lookup_host(server);
		IPport port;
		if(portno > 0) port = portno;
		else if(ftpovertls) port = lookup_service("ftps");
		else port = lookup_service("ftp");
		connect(sock, IPendpoint(addr, port));
		tls(sock, NO_CERT_CHECK, false);
		control(tls);

		cout << note_mark << "Connected" << endl;

		// If required by options, go ahead and perform
		// the TLS negotiation.
		if(ftpovertls) {
			cout << note_mark << "Starting TLS"
			     << endl;
			tls.start_tls();
			cout << note_mark << "Control TLS Established"
			     << endl;
		}

		// Get the welcome message.  Per the protocol, if we
		// get a 120 (the only possible 1xx message), we should
		// keep reading responses until we get some other kind.
		int code;
		do {
			code = response(control);
		} while(code / 100 == 1);

		// Not a 1xx or 2xx, only other possiblities are errors.
		if(code / 100 != 2)
			throw bail_out("Server did not accept "
				       "connection");

		// Establish SSL if required.
		if(starttls)
			do_request_tls
				(control, tls, datacrypt ? "P" : "C");

		// Log in to the server.
		if(login == "anonymous") {
			if(!do_login(control, login, "password"))
				throw bail_out("Anonymous login not "
					       "allowed");
		} else {
			do_user_login(control, login);
		}

		cout << note_mark << "Successfully logged in" << endl;

		// Try to move to the indicated directory, if any.
		if(argc == 2) {
			do_or_die(control, "CWD " + (string)argv[1],
				  "Directory change failed");
		}

		// Attempt to print the listing of this directory
		// in passive mode.
		do_listing(control, datacrypt ? &tls : NULL);

		// All done.
		request(control, "QUIT");
		cout << note_mark << "Complete" << endl;
		tls.stop_tls();
	}
	catch(bail_out &eq) 
	{
		// Server-reported error.  Simply print it and QUIT.
		cout << note_mark << eq.what() << endl;
		request(control, "QUIT");
		cout << note_mark << "Finished" << endl;
		tls.stop_tls();
	}
	catch(runtime_error &e)
	{
		cerr << "Error: " << e.what() << endl;
	}

	close(sock);
}
