/*
 * This test runs two threads, one to do an echo, and the other to
 * send data and check the return.
 *
 *  -u		Send UDP messages instead of using a TCP stream.  For UDP,
 *		the receiving process might not get all the messages.  Also,
 *		test may not shut down properly if the wrong messages are lost.
 *  -p pno	Local port number.  Otherwise, chosen at random.
 *  -b		Use a buffered receiver.
 *  -n bytes	How many bytes to transmit during the test.  For UDP, this
 *		is only payload, so more will be transmitted.
 *  -h holdmax  For stream, a number of bytes.  For message, a number of 
 *		messages.  Max amount the echoer will retain.  Sender makes
 *		sure to send that much extra to avoid deadlocks.  Default 3
 *		under -u, otherwise 20.
 *  -r pno	Reverse port number used briefly with UDP.  Default random.
 *  -t		Print statistics.
 *  -v		Verbose.  Print stuff.  Implies -t.
 */

/*
 * This file is part of the Cleansocks library.  It is released under the
 * Lesser GNU GPL.  See the accompanying files COPYING and COPYING.LESSER.
 */

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <iostream>
#include <string>
#include <stdexcept>
#include <stdio.h>
#include <algorithm>
#include <list>
#include <thread>
#include <mutex>
#include <array>

#include "cleansocks.h"
#include "cleanip.h"
#include "cleanbuf.h"

#ifdef WIN32
#define getpid() GetCurrentProcessId()
#endif

using namespace cleansocks;

/* Processing functions for each type of transmission. */
void tcp_child();
void tcp_parent();
void buf_child();
void buf_parent();
void udp_child();	
void udp_parent();

/* The options block.  This is thread-shared, but not changed after the
   child thread is started. */
struct opts_t {
	bool tots, verbose;
	int portno, nbytes, holdct, rev_portno;
	void (*child)(), (*parent)();
	opts_t(): verbose(false), tots(false), 
		  child(tcp_child), 
		  parent(tcp_parent), portno(0), rev_portno(0),
		  nbytes(5000), holdct(-1) {}

} opts;

static std::mutex iosync;

int main(int argc, char **argv)
{
	// Read the options.
	int ac = argc-1;
	char **av = argv+1;
	int incr = 1;
	for(; ac > 0; ac -= incr, av += incr) {
		incr = 1;
		if(av[0][0] == '-') switch(av[0][1]) {
		case 't':
			opts.tots = true;
			continue;
		case 'v':
			opts.tots = opts.verbose = true;
			continue;
		case 'u':
			opts.child = udp_child;
			opts.parent = udp_parent;
			continue;
		case 'b':
			opts.child = buf_child;
			opts.parent = buf_parent;
			continue;
		case 'p':
			opts.portno = argc > 0 ? atoi(av[1]) : 0;
			incr = 2;
			continue;
		case 'r':
			opts.rev_portno = argc > 0 ? atoi(av[1]) : 0;
			incr = 2;
			continue;
		case 'n':
			if(argc > 0) opts.nbytes = atoi(av[1]);
			incr = 2;
			continue;
		case 'h':
			if(argc > 0) opts.holdct = atoi(av[1]);
			incr = 2;
			continue;
		}
		std::cerr << "Bad flag " << av[0] << std::endl;
		exit(1);
	}

	time_t now;
	time(&now);
	srand(now ^ getpid());

	/* Choose a port number, if needed. */
	if(!opts.portno) {
		opts.portno = rand() % 50000 + 10000;
	}
	if(opts.parent == udp_parent && !opts.rev_portno) {
		opts.rev_portno = rand() % 50000 + 10000;
		if(opts.rev_portno == opts.portno) ++opts.rev_portno;
	}

	if(opts.holdct < 0) {
		if(opts.parent == udp_parent)
			opts.holdct = 3;
		else
			opts.holdct = 20;
	}

	// Start the child thread.
	std::thread childthrd(opts.child);

	// Run the parent (normal call).
	opts.parent();

	// Finish up.
	childthrd.join();
}

// Map a positive integer to some byte using some simple binary rotations.
// Not much.
unsigned char nth_byte(int n)
{
	static const int GO = ' ';
	static const int RANGE = '~' - GO;

	return (GO + ((n & 0xAB) + (n & 0x18) + (n & 0x57)) % (RANGE + 1)) & 0xff;
}

// Integer to string
std::string itos(int n)
{
	char buf[100];
	sprintf(buf, "%d", n);
	return buf;
}

std::string itoqc(int n)
{
	char buf[100];
	sprintf(buf, "'%c'", n);
	return buf;
}

// This is a pattern of sizes used for sending blocksizes, and in stream mode
// used separately for receving block sizes.
static const int SIZESIZE = 7;
static const int sizes[SIZESIZE] = { 3, 19, 49, 138, 150, 9, 74 };

// Get the next size.
int getsize(int &dx)
{
	int ret = sizes[dx++];
	dx %= SIZESIZE;
	return ret;
}

// Get the next size, but limit to a given max.  If the max is negative,
// it is treated as zero.
int getsize(int &dx, int limit)
{
	return std::min(getsize(dx), std::max(limit, 0));
}

// Helper buffered stream reader.
int bufrecv(buffered_socket & s, void *buf, int size) {
	return recvln(s, buf, size, ' ');
}

int recv_sock_already_you_stupid_compiler(TCPsocket &s, void *buf, int size) {
	return recv(s,buf,size);
}

int recv_buf_already_you_stupid_compiler(buffered_socket &s, void *buf, 
					 int size) {
	return recv(s,buf,size);
}

// Echo with funny block behavior.  Template is designed to work with
// either TCP socket or a buffered socket.
template<typename T>
void strm_child_proc(T &comm, int (*recv)(T &s, void *buf, int size))
{
	// Keeping track.
	int sendcnt = 0, rcvcnt = 0;
	int sendrot = 0, rcvrot = 5;

	// First send and receive sizes.
	int rsize = getsize(rcvrot);
	int ssize = getsize(sendrot);

	// R/S area.
	char buf[1000];
	int bufsize = 0;

	// Repeat until we've echoed all the bytes.
	bool eof = false;
	while(sendcnt < rcvcnt || !eof) {
		// Limit sending when EOF reached.
		if(eof) ssize = std::min(ssize, bufsize);

		// Are we holding enough bytes to cover the less of the send
		// size and the reserve limit?  I really want this to be just
		// bufsize >= ssize, but there's all kind of stupid details.  I
		// need the opts.holdct limit so the loop doesn't violate the
		// limit by not sending when it has too many bytes.  Also,
		// ssize is limited to the number of bytes remaining to send,
		// so it can be zero near the end of the computation.  This
		// means that a buffer of zero is big enough to send, so I
		// have a test to prevent that.
		if(bufsize > 0 && bufsize >= std::min(ssize,opts.holdct)) {
			// Yes.  Do a send.
			int tosend = std::min(ssize, bufsize);
			int nsent = send(comm, buf, tosend);

			// All the bytes should have gone.
			if(nsent != tosend)
				throw std::runtime_error
					("Short strm send: " + itos(tosend) +
					 " != " + itos(ssize));

			// Adjust buffer contents.
			bufsize -= nsent;
			if(bufsize > 0) memmove(buf, buf + nsent, bufsize);
			sendcnt += nsent;

			if(opts.verbose) {
				std::lock_guard<std::mutex> lg(iosync);
				std::cout << "Child sent " << tosend
					  << " bytes (" << sendcnt
					  << " total)" << std::endl;
			}

			// Next send size.  Limit here avoids problems at end
			// where we want to send more bytes than there are.
			ssize = getsize(sendrot, opts.nbytes - sendcnt);
		} else {
			// No.  Need to get some more bytes.  (Shouldn't
			// overflow because of the max r/c units and larger
			// buffer.  Buffer more than any two units. )
			int nrcv = recv(comm, buf + bufsize, rsize);
			if(opts.verbose) {
				std::lock_guard<std::mutex> lg(iosync);
				std::cout << "Child got " << nrcv
					  << " bytes (" << rcvcnt + nrcv
					  << ") total." << std::endl;
			}

			// If we reach EOF, need to start shutting down.
			if(nrcv == 0) {
				// Mark completion.
				eof = true;
				rsize = 0;
			} else {
				// Added to buffer.
				bufsize += nrcv;
				rcvcnt += nrcv;

				// Next receive size.
				rsize = getsize(rcvrot);
			}
		}
	}

	if(opts.tots) {
		std::lock_guard<std::mutex> lg(iosync);
		std::cout << "Child ending, sent " << sendcnt
			  << " bytes, received " << rcvcnt << " bytes."
			  << std::endl;
	}
}


template<typename T>
void strm_parent_proc(T &comm, int (*recv)(T &s, void *buf, int size))
{
	// Keeping track.
	int sendoff = 0, rcvoff = 0;
	int sendrot = 0, rcvrot = 5;

	int sendcnt = 0, rcvcnt = 0;

	// First send and receive sizes.
	int rsize = getsize(rcvrot, opts.nbytes - rcvoff);
	int ssize = getsize(sendrot, opts.nbytes - sendoff);

	// Repeat until we've transmitted the needed number of bytes.
	while(rcvoff < opts.nbytes) {
		// Have we sent enough bytes to cover the reception, plus
		// enough to cover the max that might not be forwarded
		// immediately?
		if(sendoff >= std::min(rcvoff + rsize + opts.holdct, opts.nbytes)) {
			// Yes.  Do a receive.
			std::array<char, 500> buf;
			int nrcv = cleansocks::recv(comm, buf);
			if(opts.verbose) {
				std::lock_guard<std::mutex> lg(iosync);
				std::cout << "Parent got " << nrcv
					  << " bytes" << std::endl;
			}
			rcvcnt += nrcv;

			// Verify.  Should not be zero size, and bytes
			// should be correct.
			if(nrcv == 0)
				throw std::runtime_error("Premature EOF");
			for(int i = 0; i < nrcv; ++i, ++rcvoff)
				if(buf[i] != nth_byte(rcvoff))
					throw std::runtime_error
						("Bad byte rcved at " + 
						 itos(rcvoff) + ": " +
						 itoqc(buf[i]) + " not " +
						 itoqc(nth_byte(rcvoff)));

			// Next receive size.
			rsize = getsize(rcvrot, opts.nbytes - rcvoff);
		} else {
			// No.  Make a buffer of data and send it.
			std::array<char, 500> buf;
			for(int i = 0; i < ssize; ++i)
				buf[i] = nth_byte(sendoff++);
			if(opts.verbose) {
				std::lock_guard<std::mutex> lg(iosync);
				std::cout << "Parent sending " << ssize
					  << " bytes" << std::endl;
			}
			int nsent = send(comm, buf, ssize);
			sendcnt += nsent;

			// All the bytes should have gone.
			if(nsent != ssize)
				throw std::runtime_error
					("Short strm send: " + itos(nsent) +
					 " != " + itos(ssize));
			// Next send size.
			ssize = getsize(sendrot, opts.nbytes - sendoff);
		}
	}

	if(opts.tots) {
		std::lock_guard<std::mutex> lg(iosync);
		std::cout << "Parent ending, sent " << sendcnt
			  << " bytes, received " << rcvcnt << " bytes."
			  << std::endl;
	}
}


void tcp_child()
{
	TCPsocket s;
	IPendpoint e(IPaddress::any(), opts.portno);
	bind(s,e);
	listen(s);
	TCPsocket cs = accept(s);
	//TCPsocket cs = (TCPsocket)accept(s);
	strm_child_proc<TCPsocket>(cs,recv_sock_already_you_stupid_compiler);
}

void buf_child()
{
	TCPsocket s;
	IPendpoint e(IPaddress::any(), opts.portno);
	bind(s,e);
	listen(s);
	TCPsocket cs = accept(s);
	buffered_socket bufsock(cs);
	strm_child_proc<buffered_socket>(bufsock,
					 recv_buf_already_you_stupid_compiler);
}

void tcp_parent()
{
	sleep(1);
	TCPsocket s;
	connect(s, IPendpoint("127.0.0.1", opts.portno));
	strm_parent_proc<TCPsocket>(s,recv_sock_already_you_stupid_compiler);
	close(s);
}

void buf_parent()
{
	sleep(1);
	TCPsocket s;
	connect(s, IPendpoint("127.0.0.1", opts.portno));
	buffered_socket bufsock(s);
	strm_parent_proc<buffered_socket>(bufsock,
					  recv_buf_already_you_stupid_compiler);
	close(bufsock);
}

// Insert two integers (the header) into a buffer and return a pointer to
// the following byte.
unsigned char *putints(unsigned char *buf, int a, int b)
{
	memcpy(buf, &a, sizeof a);
	buf += sizeof a;
	memcpy(buf, &b, sizeof b);
	buf += sizeof b;
	return buf;
}

// Extract two integers (the header) from a buffer into the args and return a 
// pointer to the following byte.
unsigned char *getints(unsigned char *buf, int &a, int &b)
{
	memcpy(&a, buf, sizeof a);
	buf += sizeof a;
	memcpy(&b, buf, sizeof b);
	buf += sizeof b;
	return buf;
}

// This _should_ be in the following function, but then I can't make a 
// std::list<> of them.
struct bufblock {
	int size;
	IPendpoint rmt;
	unsigned char *buf;

	bufblock(): size(0), buf(NULL) { }
	unsigned char * alloc(unsigned char *b, int s) {
		buf = new unsigned char[size = s];
		memcpy(buf, b, s);
		return buf;
	}
	void dispose() { 
		delete [] buf; 
		buf = NULL;
		size = 0;
	}
};

// Special messages: Header values are CTL and a code.
static const int CTL = -1;
static const int EOT = -2;	// End of transmission.

// Read and echo blocks.  The sizes are generated by getsize(), and are
// checked.
void udp_child()
{
	// R/S area.
	std::list<bufblock> bufs;

	int remaining = opts.nbytes;    // Remaining amt to rcv.
	int rcvoffset = 0;		// Verifier count.

	int totsent = 0;		// Total bytes sent.
	int nmsent = 0;			// Total message sent.

	bool seen_eot = false;		// Other end is done.

	UDPsocket s;
	bind(s, IPendpoint(IPaddress::any(), opts.portno));

	// Sync to the parent using a TCP handshake.
	{
		TCPsocket syncsock;
		connect(syncsock, IPendpoint("127.0.0.1", opts.rev_portno));
		close(syncsock);
	}

	// Repeat until we've echoed all the bytes.
	while(!bufs.empty() || (remaining > 0 && !seen_eot)) {
		// See what to do.  We will send if 
		//    We have received all the bytes, or
		//    We are holding the maximum number of messages, or
		//    We are holding at least one message and the coin toss
		//       comes up heads.
		if(remaining <= 0 || bufs.size() >= opts.holdct || 
		   (!bufs.empty() && (rand() & 0x10))) {
			// Send a buffer.
			if(opts.verbose) {
				std::lock_guard<std::mutex> lg(iosync);
				std::cout << "Child sending "
					  << bufs.front().size << " bytes" 
					  << std::endl;
			}
			int nsent = sendto(s, bufs.front().buf, 
					   bufs.front().size, bufs.front().rmt);

			// All the bytes should have gone.
			if(nsent != bufs.front().size)
				throw std::runtime_error
					("Short msg send: " + itos(nsent) +
					 " != " + itos(bufs.front().size));

			totsent += nsent;
			nmsent++;

			// Delete the message.
			bufs.front().dispose();
			bufs.pop_front();
		} else {
			// Read and store a buffer.
			bufs.push_back(bufblock());
			unsigned char buf[500];
			int nrcv = recvfrom(s, buf, sizeof buf, 
					    bufs.back().rmt);
			if(opts.verbose) {
				std::lock_guard<std::mutex> lg(iosync);
				std::cout << "Child got "
					  << nrcv << " bytes" << std::endl;
			}

			// Extract the header.
			int exsize, offset;
			getints(buf, exsize, offset);

			// Check for control message.
			if(exsize == CTL && offset == EOT) {
				if(opts.verbose) {
					std::lock_guard<std::mutex> lg(iosync);
					std::cout << "=== Child got EOT ==="
						  << std::endl;
				}

				seen_eot = true;

				bufs.back().dispose();
				bufs.pop_back();
				continue;
			}

			// Sanity check on size.
			if(exsize > nrcv) 
				throw std::runtime_error
					("Insane data size: " + itos(exsize) +
					 " > " + itos(nrcv));

			// Store the buffer.
			bufs.back().alloc(buf, nrcv);

			if(opts.verbose) {
				std::lock_guard<std::mutex> lg(iosync);
				std::cout << "=== Child storing " << nrcv 
					  << " bytes, " << "size " << exsize 
					  << " ===" << std::endl;
			}

			remaining -= exsize;
		}
	}

	if(remaining < 0) 
		throw std::runtime_error("Too much data rcved: " + 
					 itos(-remaining) + " extra");

	if(opts.tots) {
		std::lock_guard<std::mutex> lg(iosync);
		std::cout << "Child exiting after echoing "
			  << totsent << " bytes in " << nmsent
			  << " messages" << std::endl;
	}
}

// Send messages of random sizes up to the sending amount.
void udp_parent()
{
	int remaining = opts.nbytes;	// Remaining amt to send.
	int npend = 0;			// Number of sent messages not rcved.
	int sendrot = 0;		// Send size generator.
	int sendoff = 0;		// Send byte gen offset.

	int totrcv = 0;			// Bytes received,
	int totsend = 0;		//   and sent.

	int nmrcv = 0;			// Messages received,
	int nmsend = 0;			//   and sent.

	UDPsocket s;
	IPendpoint ept("127.0.0.1", opts.portno);

	// Sync with the child using a TCP handshake.
	{
		TCPsocket syncsock;
		bind(syncsock, IPendpoint(IPaddress::any(), opts.rev_portno));
		listen(syncsock);
		TCPsocket cs = accept(syncsock);
		close(cs);
		close(syncsock);
	}

	// Repeat until we've transmitted the needed number of bytes.
	while(remaining > 0 || npend > 0) {
		// Do a rececive if:
		//    We have sent all the bytes needed, or
		//    The number of pending messages exceeds the hold count, and
		//         the coin comes up heads.
		if(remaining == 0 || (npend > opts.holdct && (rand() & 0x10))) {
			// Yes.  Do a receive.
			unsigned char buf[500];
			int nrcv = recvfrom(s, buf, sizeof buf);

			// Extract the header data.
			int gensize, rcvoff;
			unsigned char *bscan = getints(buf, gensize,rcvoff);

			// One less pending message, more total bytes received.
			--npend;
			totrcv += nrcv;
			nmrcv++;

			if(opts.verbose) {
				std::lock_guard<std::mutex> lg(iosync);
				std::cout << "Parent got "
					  << nrcv << " bytes, " << gensize
					  << " payload, key = " << rcvoff
					  << std::endl;
			}

			// Check the size.
			int expsize = gensize + sizeof gensize + sizeof rcvoff;
			if(nrcv != expsize)
				throw std::runtime_error
					("Bad msg rcv size: " + itos(nrcv) +
					 " != " + itos(expsize));

			// Verify bytes.
			for(int i = 0; i < gensize; ++i, ++rcvoff, ++bscan)
				if(*bscan != nth_byte(rcvoff))
					throw std::runtime_error
						("Bad byte rcved at " + 
						 itos(rcvoff) + ": " +
						 itoqc(*bscan) + " not " +
						 itoqc(nth_byte(rcvoff)));
		} else {
			// No.  Make a buffer of data and send it.  The
			// size and content counters used to generate the
			// message are added to the front.
			unsigned char buf[500];
			int gensize = getsize(sendrot, remaining);
			unsigned char *bscan = putints(buf, gensize, sendoff);
			for(int i = 0; i < gensize; ++i)
				*bscan++ = nth_byte(sendoff++);
			int sendsize = bscan - buf;
			int nsent = sendto(s, buf, sendsize, ept);
			totsend += nsent;
			nmsend++;

			// All the bytes should have gone.
			if(nsent != sendsize)
				throw std::runtime_error
					("Short msg send: " + itos(nsent) +
					 " != " + itos(sendsize));

			if(opts.verbose) {
				std::lock_guard<std::mutex> lg(iosync);
				std::cout << "+++ Parent sent " << nsent 
					  << " bytes, " << gensize 
					  << " payload, key = "
					  << sendoff - gensize << std::endl;
			}

			remaining -= gensize;
			++npend;
		}
	}

	// Send a term.
	{
		int endmsg[] = { CTL, EOT };
		for(int i = 3; i--; )
			sendto(s, (unsigned char *)endmsg, sizeof endmsg, ept);	
	}

	close(s);

	if(opts.tots) {
		std::lock_guard<std::mutex> lg(iosync);
		std::cout << "Parent ending, sent " << totsend << " bytes in "
			  << nmsend << " messages, got " << totrcv << " bytes "
			  << "in " << nmrcv << " messages." << std::endl;
	}
}
