/*
 * Run with a single HTTP URL on the command line.  Will attempt to fetch it
 * as HTTP/1.0 and print the results.  URL parsing is none-too-sophisticated,
 * so don't expect much.  
 */
	
/*
 * This file is part of the Cleansocks library.  It is released under the
 * Lesser GNU GPL.  See the accompanying files COPYING and COPYING.LESSER.
 */
	
#include <stdlib.h>
	
#include <iostream>
#include <string>
#include <stdexcept>

#include "cleansocks.h"
#include "cleanip.h"
#include "cleantlsc.h"

using namespace std;
using namespace cleansocks;

void doit(string url, cleantlsc_flags tflags, std::string bogus)
{
	// Check and remove the leading http://
	bool tls = false;

	int len = 8;
	if(url.length() >= len && url.substr(0,len) == "https://")
		tls = true;
	else {
		len = 7;
		if(url.length() < len || url.substr(0,len) != "http://")
			throw runtime_error(url + " does not start with "
					    "\"http://\" or \"https://\"");
	}
	url.erase(0,len);
	
	// Isolate the host name.
	int port = tls ? 443 : 80;
	string hostname = url;
	string path = "/";

	// Isolate the host name, if it's not the entire URL.
	int hend = url.find_first_of("/:");
	if(hend != string::npos) {
		hostname = url.substr(0,hend);

		// Extract a port if present.
		if(url[hend] == ':') {
			port = atoi(url.substr(hend+1).c_str());
			hend = url.find("/", hend+1);
		}
	}

	// The rest of the URL is the path, or /.
	if(hend == string::npos)
		path = "/";
	else 
		path = url.substr(hend);

	cout << "host = " << hostname << ", port = " << port
	     << ", path = " << path << (tls ? " [secure]" : "")
	     << endl;

	// Look up the host and try to connect.
	IPaddress a = lookup_host(hostname);
	IPport p = port;
	TCPsocket cs;
	connect(cs,IPendpoint(a,p));

	IPendpoint local;
	getsockname(cs, local);
	std::cout << "Connected to " << a << ":" << p
		  << " from local " << local << endl;

	client_tls_socket s(cs,tflags,bogus==""?hostname:bogus,tls);

	// Write an HTTP request to the server.
	send(s, "GET "+path+" HTTP/1.0\r\n");
	send(s, "Host: "+hostname+"\r\n");
	send(s, static_cast<string>("User-Agent: CleanSocksTester\r\n\r\n"));

	// Read and print the response.  Use a small buffer so it will
	// usually need multiple calls.
	cout << endl;
	char buf[100];
	int n;
	while((n = recv(s, buf, sizeof buf)) != 0)
		cout.write(buf, n);

	close(s);
}

int main(int argc, char **argv)
{
	if(argc < 2) {
		cerr << "Usage: " << argv[0] << "[flags] url" << endl;
		exit(2);
	}

	std::string url = argv[--argc];
	
	// Flags, mostly for some tls verification things.  These are
	// ignored for a plain http: URL
	cleantlsc_flags tflags(0);
	std::string bogus = "";
	--argc; ++argv;

	while(argc > 0) {
		std::string a = *argv;
		if(a == "NO_CERT_CHECK") tflags |= NO_CERT_CHECK;
		else if(a == "SELF_OK") tflags |= SELF_OK;
		else if(a == "EXPIRED_OK") tflags |= EXPIRED_OK;
		else if(a == "NO_CERT_OK") tflags |= NO_CERT_OK;
		else if(a == "NO_SNI") tflags |= NO_SNI;
		else if(a == "NO_HOST_CHK") tflags |= NO_HOST_CHK;
		else bogus = a;

		--argc; ++argv;
	}
	
	try {
		doit(url, tflags, bogus);
	} catch(socket_db_error &e) {
		cout << "Socket db error: " << e.what() << endl;
	}
	catch(socket_tls_error &e) {
		cout << "TLS error: " << e.what() << endl;
		socket_tls_error::error_stack();
	}
	catch(socket_error &e) {
		cout << "Socket error: " << e.what() << endl;
	}
	catch(runtime_error &e) {
		cout << "Runtime error: " << e.what() << endl;
	}
}

